#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
File for reading and setting configurations for the whole library.
I will not pretend this config file is a good way to do things.
But it is flexible, and works so far.
I make no guarantees that this will not destroy your computer.

Example config file:

``` ini
[Paths]
; Root directories, under which filings/cached tarballs will be found
; There is no % interpolation

; CACHE_FEED indicates whether the feed should be cached/searched locally
CACHE_FEED = True

; FILING_ROOT is the root of the extracted filings
FILING_ROOT=/data/bulk/data/edgar/filings/

; FEED_CACHE_ROOT is the root of the compressed daily feed files from EDGAR
FEED_CACHE_ROOT=/data/bulk/data/edgar/raw_from_edgar/compressed_daily_feeds/

; CACHE_INDEX indicates whether the index should be cached/searched locally
CACHE_INDEX = True

; INDEX_ROOT is the root of the extracted index tab-delimited files
INDEX_ROOT=/data/bulk/data/edgar/indices/

; INDEX_CACHE_ROOT is the root of the
INDEX_CACHE_ROOT=/data/bulk/data/edgar/raw_from_edgar/indices/

; FILING_PATH_FORMAT is the string to be .format-ed with the CIK and ACCESSION of the filing
; Don't put injection attacks here. That would be bad.
; Maximum length is 250 characters.
; Format string is formatted as an f-string (see docs), therefore slicing is possible.
; Available variables are:
;    cik (int)
;    cik_str (=f'{cik:010d}')
;    accession (20 character format with dashes)
;    and accession18 (18 characters of only digits with dashes removed)
; Examples:
; FILING_PATH_FORMAT={accession[11:13]}/{accession}.nc
;         Would result in --> FILING_ROOT/95/0001005463-95-000003.nc
;         This is useful for accession-only lookups (which is nice because multiple CIKs can file the same accession)
;
; FILING_PATH_FORMAT={cik_str[0:2]}/{cik_str[2:4]}/{cik_str[4:6]}/{cik_str[6:8]}/{cik_str[8:10]}/{accession}.txt
;         Would result in --> FILING_ROOT/00/01/00/54/63/0001005463-95-000003.txt
;         This uses CIK to break up filings, resulting in < 100 entries per directory. One problem is multiple CIKs
;           can file the same accession, meaning you have to either copy the same accession filing to multiple dirs
;
; FILING_PATH_FORMAT={accession[:4]}/{accession[4:7]}/{accession[7:10]}/{accession[11:13]}/{accession[14:17]}/{accession[17:]}/{accession}.nc
;         Would result in --> FILING_ROOT/1234/567/890/12/123/456/1234567890-12-123456.txt
;         This is useful for only accession lookups (no CIKs) but also < 1000 entries per directory
;
FILING_PATH_FORMAT={accession[11:13]}/{accession}.nc

; Filename format for caching FEED compressed files from EDGAR
; String is passed .format(date=datetime object) of the date of the feed
FEED_CACHE_PATH_FORMAT={date:%Y%m%d}.nc.tar.gz

; Filename format for caching INDEX compressed files from EDGAR
; Available data are: date (datetime object), year, and quarter (both ints)
INDEX_CACHE_PATH_FORMAT=full_index_{year}_Q{quarter}.gz

[Downloader]
; Downloader specific settings
KEEP_ALL=True
KEEP_REGEX=

; User Agent for downloading, to keep the SEC happy
USER_AGENT=pyedgar feed download by YOUREMAIL@example.com, from code at https://github.com/gaulinmp/pyedgar

[Index]
; Index file settings
INDEX_DELIMITER=\t
; Index file extension
; If you want to compress the index files, change INDEX_EXTENSION to tab.gz
INDEX_EXTENSION=tab.gz
```

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""


# STDlib imports
import os
import re
import logging
import tempfile
import configparser
import datetime as dt
from itertools import product, starmap

# 3rd party imports

_logger = logging.getLogger(__name__)


#  ██████╗ ██████╗ ███╗   ██╗███████╗██╗ ██████╗     ███████╗██╗██╗     ███████╗
# ██╔════╝██╔═══██╗████╗  ██║██╔════╝██║██╔════╝     ██╔════╝██║██║     ██╔════╝
# ██║     ██║   ██║██╔██╗ ██║█████╗  ██║██║  ███╗    █████╗  ██║██║     █████╗
# ██║     ██║   ██║██║╚██╗██║██╔══╝  ██║██║   ██║    ██╔══╝  ██║██║     ██╔══╝
# ╚██████╗╚██████╔╝██║ ╚████║██║     ██║╚██████╔╝    ██║     ██║███████╗███████╗
#  ╚═════╝ ╚═════╝ ╚═╝  ╚═══╝╚═╝     ╚═╝ ╚═════╝     ╚═╝     ╚═╝╚══════╝╚══════╝

PREFERRED_CONFIG_DIRECTORIES = [
    os.path.abspath(os.curdir),
    os.path.expanduser("~/.config/pyedgar"),
    os.path.expanduser("~/AppData/Local/pyedgar"),
    os.path.expanduser("~/AppData/Roaming/pyedgar"),
    os.path.expanduser("~/Library/Preferences/pyedgar"),
    os.path.expanduser("~/.config"),
    os.path.expanduser("~"),
    os.path.expanduser("~/Documents"),
]


def get_config_file(extra_dirs=None):
    """Searches the expected config paths and names, returning the first found.

    First, try and load directory location from environmental variable `PYEDGAR_CONF`
    if it points to a specific file.

    Directories searched, in order:

    #. `os.environ['PYEDGAR_CONF']` (could be directory path as well)
    #. `.` (current dir)
    #. `~/.config/pyedgar/`
    #. `~/AppData/Local/pyedgar/`
    #. `~/AppData/Roaming/pyedgar/`
    #. `~/Library/Preferences/pyedgar/`
    #. `~/.config/`
    #. `~/`
    #. `~/Documents/`

    For each of those directories, the following file names are looked for:

    #. `pyedgar.conf`
    #. `.pyedgar`
    #. `pyedgar.ini`

    The first one of these combinations that is found is read and returned.
    """

    try:
        with open(os.environ["PYEDGAR_CONF"], "r") as fh:
            config_txt = fh.read()
            if config_txt:
                return os.environ["PYEDGAR_CONF"]
    except (KeyError, FileNotFoundError):
        pass

    _names = ["pyedgar.conf", ".pyedgar", "pyedgar.ini"]

    _dirs = [
        os.environ.get("PYEDGAR_CONF", "."),  # env variable might point to dir not file
        *PREFERRED_CONFIG_DIRECTORIES,
    ]

    if extra_dirs:
        if isinstance(extra_dirs, str):
            _dirs.append(extra_dirs)
        else:
            _dirs.extend(extra_dirs)

    config_txt = None
    for fpath in starmap(os.path.join, product(_dirs, _names)):
        try:
            with open(fpath, "r") as fh:
                config_txt = fh.read()
                if config_txt:
                    return fpath
        except (IOError, FileNotFoundError):
            pass

    # getting here means we didn't find the file
    return None


_tmp_dir = os.path.join(tempfile.gettempdir(), "pyedgar")

_defaults = {
    "FILING_ROOT": os.path.join(_tmp_dir, "filings"),
    "FEED_CACHE_ROOT": os.path.join(_tmp_dir, "compressed_daily_feeds"),
    "CACHE_FEED": "False",
    "INDEX_ROOT": os.path.join(_tmp_dir, "indices"),
    "INDEX_CACHE_ROOT": os.path.join(_tmp_dir, "indices"),
    "CACHE_INDEX": "False",
    "FILING_PATH_FORMAT": "{accession[11:13]}/{accession}.nc",
    "FEED_CACHE_PATH_FORMAT": "sec_daily_{date:%Y-%m-%d}.tar.gz",
    "INDEX_CACHE_PATH_FORMAT": "full_index_{year}_Q{quarter}.gz",
    "KEEP_ALL": "True",
    "KEEP_REGEX": "",
    "INDEX_DELIMITER": "\t",
    "INDEX_EXTENSION": "tab",
    "USER_AGENT": "pyedgar feed download by YOUREMAIL@example.com, from code at https://github.com/gaulinmp/pyedgar",
}

CONFIG_FILE = get_config_file()
_logger.info("Config file to be loaded: %r", CONFIG_FILE)

CONFIG_OBJECT = configparser.ConfigParser(interpolation=None, defaults=_defaults)
try:
    CONFIG_OBJECT.read(CONFIG_FILE)
    _logger.info(
        "Loaded config file from %r. \n\n" "ALERT!!!! FILING_PATH_FORMAT is %r.\n",
        CONFIG_FILE,
        CONFIG_OBJECT.get("Paths", "FILING_PATH_FORMAT", fallback=None),
    )
except TypeError:
    # Type error means that we tried to read from None file
    _logger.info("Error reading config file: %r", CONFIG_FILE)
    # Come on python... how does a nonexistent section not drop through to DEFAULT?!
    for sec in ("Paths", "Downloader", "Index"):
        CONFIG_OBJECT.add_section(sec)
except Exception:
    _logger.exception("Error reading config file: %r", CONFIG_FILE)
    raise


#  ██████╗ ██████╗ ███╗   ██╗███████╗████████╗ █████╗ ███╗   ██╗████████╗███████╗
# ██╔════╝██╔═══██╗████╗  ██║██╔════╝╚══██╔══╝██╔══██╗████╗  ██║╚══██╔══╝██╔════╝
# ██║     ██║   ██║██╔██╗ ██║███████╗   ██║   ███████║██╔██╗ ██║   ██║   ███████╗
# ██║     ██║   ██║██║╚██╗██║╚════██║   ██║   ██╔══██║██║╚██╗██║   ██║   ╚════██║
# ╚██████╗╚██████╔╝██║ ╚████║███████║   ██║   ██║  ██║██║ ╚████║   ██║   ███████║
#  ╚═════╝ ╚═════╝ ╚═╝  ╚═══╝╚══════╝   ╚═╝   ╚═╝  ╚═╝╚═╝  ╚═══╝   ╚═╝   ╚══════╝

# Paths to filings and indices
FILING_ROOT = CONFIG_OBJECT.get("Paths", "FILING_ROOT")
FEED_CACHE_ROOT = CONFIG_OBJECT.get("Paths", "FEED_CACHE_ROOT")
INDEX_ROOT = CONFIG_OBJECT.get("Paths", "INDEX_ROOT")
INDEX_CACHE_ROOT = CONFIG_OBJECT.get("Paths", "INDEX_CACHE_ROOT")
# expand user dir if present
if '~' in FILING_ROOT:
    FILING_ROOT = os.path.expanduser(FILING_ROOT)
if '~' in FEED_CACHE_ROOT:
    FEED_CACHE_ROOT = os.path.expanduser(FEED_CACHE_ROOT)
if '~' in INDEX_ROOT:
    INDEX_ROOT = os.path.expanduser(INDEX_ROOT)
if '~' in INDEX_CACHE_ROOT:
    INDEX_CACHE_ROOT = os.path.expanduser(INDEX_CACHE_ROOT)

# Path format
FILING_PATH_FORMAT = CONFIG_OBJECT.get("Paths", "FILING_PATH_FORMAT")
FEED_CACHE_PATH_FORMAT = CONFIG_OBJECT.get("Paths", "FEED_CACHE_PATH_FORMAT")
INDEX_CACHE_PATH_FORMAT = CONFIG_OBJECT.get("Paths", "INDEX_CACHE_PATH_FORMAT")

# Filings cache settings
CACHE_FEED = CONFIG_OBJECT.getboolean("Paths", "CACHE_FEED")
KEEP_ALL = CONFIG_OBJECT.getboolean("Downloader", "KEEP_ALL")
KEEP_REGEX = CONFIG_OBJECT.get("Downloader", "KEEP_REGEX")
USER_AGENT = CONFIG_OBJECT.get("Downloader", "USER_AGENT")

# Index cache settings
CACHE_INDEX = CONFIG_OBJECT.getboolean("Paths", "CACHE_INDEX")
INDEX_DELIMITER = CONFIG_OBJECT.get("Index", "INDEX_DELIMITER")
INDEX_EXTENSION = CONFIG_OBJECT.get("Index", "INDEX_EXTENSION").lstrip('.')

if INDEX_DELIMITER.lower() in ("\t", "\\t", "tab", "\\\t", "\\\\t"):
    INDEX_DELIMITER = "\t"


# ██████╗  █████╗ ████████╗██╗  ██╗███████╗
# ██╔══██╗██╔══██╗╚══██╔══╝██║  ██║██╔════╝
# ██████╔╝███████║   ██║   ███████║███████╗
# ██╔═══╝ ██╔══██║   ██║   ██╔══██║╚════██║
# ██║     ██║  ██║   ██║   ██║  ██║███████║
# ╚═╝     ╚═╝  ╚═╝   ╚═╝   ╚═╝  ╚═╝╚══════╝


def format_filing_path(**kwargs):
    """
    "If at first you don\'t succeed, don\'t try skydiving."

    That makes about as much sense as a docstring as the workaround below.
    It\'s... horrifying what I\'ve done here.
    So lemme splain the what and why.
    We want a conf file that is useful for paths and whatever format you want to keep your files in on your local copy of EDGAR.
    So to facilitate this, let\'s allow an f-string type FILING_PATH_FORMAT input.
    So how to evaluate it? We can\'t use ``FILING_PATH_FORMAT.format(**kwargs)`` because then you can\'t do cik[:5] for sub-folders.
    Well here comes eval to the rescue! I know, eval-ing user-provided strings is TERRIBLE practice.
    Really TERRIBLE.
    Whatever. It works. I limit it to 250 characters anyway, so your injection has to be shorter than that.
    I'm pretty sure rm -rf / is longer than 250 characters.

    LEGAL DISCLAIMER: Know where you config file is loading from. This is most definitely an attack vector.

    Anyway, back to the code. The trick here is that I don\'t know what format the input passed in will be.
    So I just take all the input (all kwargs) and load it into local.
    Turns out this only works because of the eval statement?
        (See: https://stackoverflow.com/a/8028785/1959876)

    Cool. So we put everything you pass in into the locals, to be accessed by the dynamically eval-ed f-string.

    Last bit is for some sanity. I force the existence of the following variables:
        cik: in int form
        cik_str: in string form, =``f'{cik:010d}'``
        accession: 20 character format with dashes
        accession18: 18 characters of only digits with dashes removed
    So you can safely define the format string using these for variables, and if you're feeling adventuresome, you can add your custom variable to the format string, then pass it into this method and it should work.
    Should. Probably won't. Just stick to CIK/Accession.

    Lastly... sorry for all this.
    """
    # Get that thing from above.
    global FILING_PATH_FORMAT

    # If only someone who knew how to program would write this instead of me.
    locals().update(kwargs)

    try:
        cik = int(kwargs.get("cik", 0))
        cik_str = f"{cik:010d}"
    except (ValueError, TypeError):
        cik = 0
        cik_str = 10 * "0"

    accession = kwargs.get("accession", "9090909090-90-909090")
    accession18 = accession.replace("-", "")

    return eval(f"""f'{FILING_PATH_FORMAT[:250]}'""")


def format_feed_cache_path(datetime_in):
    """
    Formats feed cache path on a given date (from date-time input).
    """
    # Get that thing from above.
    global FEED_CACHE_PATH_FORMAT

    if isinstance(datetime_in, int):
        datetime_in = dt.date.fromordinal(datetime_in)

    # Pass datetime as both the first arg and as date=,
    # in case the config file assumes positional data
    _logger.debug("FEED_CACHE_PATH_FORMAT:%r.format(date=%s)",
                  FEED_CACHE_PATH_FORMAT, datetime_in)

    return FEED_CACHE_PATH_FORMAT.format(datetime_in, date=datetime_in)


def format_index_cache_path(datetime_or_yearQN_str):
    """
    Formats index cache path on a given date.
    Allows a date-time input, or a year-quarter string in the form yyyyQq
    """
    # Get that thing from above.
    global INDEX_CACHE_PATH_FORMAT

    def get_yr_qtr_from_str(yq_string, *, yr_re=re.compile(r"([12]\d{3})Q?([1234])", re.I)):
        """Get year/qtr from string"""
        yq = yr_re.search(yq_string)
        return tuple(int(x) for x in yq.groups()) if yq else (1990, 0)

    try:
        year = datetime_or_yearQN_str.year
        qtr = (datetime_or_yearQN_str.month - 1) // 3 + 1
    except AttributeError:
        year, qtr = get_yr_qtr_from_str(datetime_or_yearQN_str)

        datetime_or_yearQN_str = dt.datetime(year, qtr * 3 - 2, 1)

    return INDEX_CACHE_PATH_FORMAT.format(date=datetime_or_yearQN_str, year=year, quarter=qtr)


def get_filing_path(**kwargs):
    """
    Return path for feed tar file.
    This implementation requires a datetime object.
    """
    global FILING_ROOT

    return os.path.join(FILING_ROOT, format_filing_path(**kwargs))


def get_feed_cache_path(datetime_in):
    """
    Return path for feed tar file.
    This implementation requires a datetime object.
    """
    global FEED_CACHE_ROOT

    return os.path.join(FEED_CACHE_ROOT, format_feed_cache_path(datetime_in))


def get_index_cache_path(datetime_or_yearQN_str):
    """
    Returns the full path for index tar file.
    Requires a datetime object or string of format 2000Q4.
    """
    global INDEX_CACHE_ROOT

    return os.path.join(INDEX_CACHE_ROOT, format_index_cache_path(datetime_or_yearQN_str))



# -*- coding: utf-8 -*-
"""
pyEDGAR SEC data library: Utilities package
======================================================

These utilities represent lower level functionality, used by the main modules.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import re as __re
import datetime as __dt
import logging as __logging

__logger = __logging.getLogger(__name__)


def get_cik_acc(*args, additional_extract=None, accession_pattern=__re.compile("\d{10}-?\d{2}-?\d{6}"), **kwargs):
    """
    I found myself wanting to accept flexible cik/acc inputs, and wrote this
    input parsing over and over. So finally we'll DRY and centralize.

    Examples:
        Can be called in any of the following ways::

            cik,acc = 1750, '1234567890-12-123456'
            get_cik_acc(cik, acc)
            get_cik_acc(acc, cik)
            get_cik_acc(namedtuple('foo', ['cik', 'accession'])(cik=cik, accession=acc))
            get_cik_acc({'cik':cik, 'accession':acc})
            get_cik_acc({'cik':cik, 'accession':acc, 'gvkey': 1, 'datadate':dt.date(2020,1,1)})
            get_cik_acc({'cik':cik, 'accession':acc, 'gvkey': 1, 'datadate':dt.date(2020,1,1)}, additional_extract='gvkey datadate'.split())
            get_cik_acc(**{'cik':cik, 'accession':acc, 'gvkey': 1, 'datadate':dt.date(2020,1,1)}, additional_extract='datadate')
            get_cik_acc(**{'cik_from':cik, 'accession_from':acc}, additional_extract='cik_from accession_from'.split())


    Arguments:
        positional arguments: will grab the first cik and acccession looking values, put the rest in 'args' key
        additional_extract (str, list): one (str) or more (list) additional keywords to extract from inputs
        kwargs: will overwrite any positional arg matches, put non-looked for

    Returns:
        tuple: Tuple of (cik, accession, etc?)
    """
    _ret = {}
    _unused_args = []
    _unused_kwargs = {}

    if additional_extract is None:
        additional_extract = []
    elif isinstance(additional_extract, str):
        additional_extract = [additional_extract]

    def _get_thing(thing, box_o_things):
        try:
            return getattr(box_o_things, thing)
        except AttributeError:
            pass
        try:
            return box_o_things.get(thing, None)
        except AttributeError:
            pass
        return None

    if len(args):
        # Here, look for cik & acc, put the rest into
        for arg in args:
            if arg is None:
                continue

            if isinstance(arg, str) and accession_pattern.search(arg):
                # Definitively an accession
                if "accession" in _ret:
                    _unused_args.append(arg)
                else:
                    _ret["accession"] = accession_pattern.search(arg).group(0)
            elif "cik" in _ret:  # If we already have a cik, then don't overwrite it
                _unused_args.append(arg)
            elif isinstance(arg, (str, int, float)):  # ciks are just numbers
                try:
                    # int of float because pandas might convert cik to float
                    _ret["cik"] = int(float(arg))
                except ValueError:  # no TypeError beause of isinstance match above
                    _unused_args.append(arg)
            else:
                # At this point, if cik/acc is num/str format, we handled it.
                # So we just have dicts/objects to check, then bail
                _val = _get_thing("cik", arg)
                if _val is not None:
                    try:
                        # int of float because pandas might convert cik to float
                        _ret["cik"] = int(float(_val))
                    except (ValueError, TypeError):  # no TypeError beause of isinstance match above
                        pass
                _val = _get_thing("accession", arg)
                if _val is not None:
                    if isinstance(_val, str) and accession_pattern.search(_val):
                        # Definitively an accession
                        if "accession" not in _ret:
                            _ret["accession"] = accession_pattern.search(_val).group(0)

                for addtl_val in additional_extract:
                    _val = _get_thing(addtl_val, arg)
                    if _val is not None:
                        _ret[addtl_val] = _val
                # We don't know if we used them all, so tack the dict/obj on to the 'unused'
                _unused_args.append(arg)

    if len(kwargs):
        try:
            # int of float because pandas might convert cik to float
            _ret["cik"] = int(float(kwargs["cik"]))
        except (ValueError, TypeError, KeyError):
            # Wasn't the right type/parseable, or no cik key
            pass

        _val = kwargs.get("accession", None)
        if isinstance(_val, str) and accession_pattern.search(_val):
            # Definitively an accession
            _ret["accession"] = accession_pattern.search(_val).group(0)

        for addtl_val in additional_extract:
            _val = kwargs.get(addtl_val, None)
            if _val is not None:
                _ret[addtl_val] = _val

        for _key, _val in kwargs.items():
            if _key not in _ret:
                _unused_kwargs[_key] = _val

    if len(_unused_args):
        _ret["args"] = _unused_args
    if len(_unused_kwargs):
        _ret["kwargs"] = _unused_kwargs
        for _key, _val in _unused_kwargs.items():
            if _key not in _ret:
                _ret[_key] = _val

    return _ret


def parse_date_input(
    date,
    default=None,
    dt_re=__re.compile(r"([12]\d{3})[^0-9]+(\d\d?)[^0-9]+(\d\d?)"),
    dtnodelim_re=__re.compile(r"([12]\d{3})(\d\d)(\d\d)"),
    qtr_re=__re.compile(r"([12]\d{3})[Qq]([1234])"),
    yr_re=__re.compile(r"[12]\d{3}"),
):
    """
    Casts something to a date, defaulting to yesterday if nothing is passed in.
    The something can be:

        date: dt.date(2000, 1, 1)
        year (int or string): 2000 or '2000'
        date string: 20001231 or 2000-1-5 (the former must be MM, the latter can omit leading 0s)
        quarter string: 2001Q3

    Arguments:
        date (str, date, int, None): Input that makes sense cast to a date. Years will be January 1st,
            quarters will be on the 1st day of the 1st month in the quarter.
    """
    if default is None or not hasattr(default, "year"):
        default = __dt.date.fromordinal(__dt.date.today().toordinal() - 1)
    _d = None  # keep _d None until it's date, easier for ifs below

    if date is None:
        _d = default
    elif isinstance(date, str):
        _ymd = dt_re.search(date) or dtnodelim_re.search(date)
        if _ymd:
            _d = __dt.date(*map(int, _ymd.groups()))

        if _d is None and qtr_re.search(date):
            _y, _q = map(int, qtr_re.search(date).groups())
            _d = __dt.date(_y, _q * 3 - 2, 1)

        if _d is None and yr_re.search(date):
            _d = __dt.date(int(yr_re.search(date).group(0)), 1, 1)
    else:  # it could be a date, try that
        try:
            _d = __dt.date(date.year, date.month, date.day)
        except AttributeError:
            pass

    if _d is None:  # well, it's not none, date, or various strings, try int
        try:
            _d = __dt.date(int(date), 1, 1)
        except (TypeError, ValueError):
            pass

    # We've handled none, dates, int, and strings, so at this point we gotta give up
    if not hasattr(_d, "year"):
        raise ValueError("Input format not recognized: {}".format(date))

    return _d


def get_quarter(datetime_in):
    """
    Return the quarter (1-4) based on the month.
    Input is either a datetime object (or object with month attribute) or the month (1-12).
    """
    try:
        return int((datetime_in.month - 1) / 3) + 1
    except AttributeError:
        return int((datetime_in - 1) / 3) + 1


def iterate_dates(from_date, to_date=None, period="daily", skip_weekends=True, inclusive=True):
    """
    Iterates over a date range at a given 'periodicity', where the period is represented by the first occuring date.
    So Q1, 2000 would be `dt.date(2000, 1, 1)`.
    Can skip weekends, but does not know about holidays.

    Arguments:
        from_date (str, datetime, int): Can be datetime, year (int or string), string of format YYYYMMDD
            or YYYY-M-D, or string of format YYYYq[1234].
        to_date (str, datetime, int, None): Can be datetime, year (int or string), string of format YYYYMMDD
            or YYYY-M-D, or string of format YYYYq[1234]. Default: yesterday.
        period (str): Periodicity of dates yielded, either `yearly`, `quarterly`, or `daily`. Default to daily.
        skip_weekends (bool): Flag for whether to skip returning weekends (sat/sun). Default: True.
        inclusive (bool): Flag for whether to include the to_date (or the quarter/year of to_date). Default: True.

    Returns:
        (datetime): Yields datetime objects
    """
    period = str(period).lower()[0]
    _from = parse_date_input(from_date)
    _to = parse_date_input(to_date)

    if _from > _to:  # greater than means after in date math
        _from, _to = _to, _from

    if period == "y":
        for i_yr in range(_from.year, _to.year + inclusive):
            yield __dt.date(i_yr, 1, 1)
    elif period == "q":
        _yqfrom = _from.year * 4 + (_from.month - 1) // 3
        _yqto = _to.year * 4 + (_to.month - 1) // 3
        for i_qtr in range(_yqfrom, _yqto + inclusive):
            # Now reverse the yr * 4 + qtr math above.
            # I solved this mathematically, without error.
            # Definitely not trial and error in a notebook.
            yield __dt.date(i_qtr // 4, (i_qtr % 4) * 3 + 1, 1)
    else:
        for i_date in range(_from.toordinal(), _to.toordinal() + inclusive):
            i_date = __dt.date.fromordinal(i_date)
            if skip_weekends and i_date.weekday() >= 5:
                continue
            yield i_date



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
High level downloading functionality, using lower-level download routines from `pyedgar.utilities`


:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import re
import os
import logging
import datetime as dt

# 3rd party imports
try:
    from tqdm import tqdm
except ImportError:
    # If tqdm isn't available, just pass through first argument
    def tqdm(_iterable, *args, **kwargs):
        return _iterable


# Module Imports
from pyedgar import config
from pyedgar import utilities
from pyedgar.utilities import edgarcache
from pyedgar.utilities import edgarweb
from pyedgar.utilities import indices

# from pyedgar.utilities import localstore

# Local logger
_logger = logging.getLogger(__name__)


def main(start_date=None, last_n_days=30, get_indices=False, get_feeds=False, use_curl_to_download=None):
    """
    Download feeds and indices. Feeds will be downloaded for `start_date` through yesterday,
    or for the past `last_n_days` days.

    Examples:
        This will download/extract last 30 days of forms and all indices:

            ```python -m pyedgar.downloader -i -d --last-n-days 30```

        This will download and extract the last 7 days of forms:

            ```python -m pyedgar.downloader -d --last-n-days 7```

    Args:
        start_date (date): Date to start extraction of feeds from. When empty, defaults to today() - last_n_days
        last_n_days (int): If start_date is missing, extract this number days before today. Default: 30
        get_indices (bool): Flag to download and extract index files. Default: False
        get_feeds (bool): Flag to download daily feed files since `start_date` or for `last_n_days`. Default: False
        use_curl_to_download (bool, None): Flag to use cURL subprocess instead of `requests` library. If None,
            will check for and use cURL if it exists. Default: None
    """
    if use_curl_to_download is None:
        use_curl_to_download = edgarweb.has_curl()

    if start_date is None:
        start_date = dt.date.fromordinal(dt.date.today().toordinal() - last_n_days)
    else:
        start_date = utilities.parse_date_input(start_date)

    if get_feeds:
        cacher = edgarcache.EDGARCacher(
            keep_form_type_regex=re.compile(config.KEEP_REGEX, re.I) if not config.KEEP_ALL else None,
            check_cik="cik" in config.FILING_PATH_FORMAT,
            use_requests=not use_curl_to_download,
        )
        _logger.info("Downloading since {:%Y-%m-%d}...".format(start_date))

        num_dates = len([1 for _ in utilities.iterate_dates(start_date)])
        for i_date in tqdm(utilities.iterate_dates(start_date), total=num_dates, desc="Downloading Feeds"):
            # download one date, so we can track progress with TQDM
            cacher.extract_daily_feeds(i_date, to_date=i_date, download_first=True, overwrite=False)

        _logger.info("Done downloading feeds on {:%Y-%m-%d}...".format(i_date))

    if get_indices:
        _logger.info("Downloading and extracting indices")
        # the last index file we find is probably not 'complete' because it was downloaded during the month maybe.
        # Let's make sure that isn't the case
        max_date, last_index = dt.date(1995, 1, 1), None
        for i_date in utilities.iterate_dates(1995, period="quarterly"):
            _idx = config.get_index_cache_path(i_date)
            if os.path.exists(_idx) and i_date > max_date:
                max_date, last_index = i_date, _idx
        if last_index is not None:
            _logger.info("Removing last of the old index caches: %s", last_index)
            os.remove(last_index)

        index_maker = indices.IndexMaker(use_tqdm=True, use_requests=not use_curl_to_download)
        index_maker.extract_indexes()
        _logger.info("Done downloading and extracting indices")


def print_cache_status():
    for i_date in reversed(list(utilities.iterate_dates(1995))):
        _feedfile = config.get_feed_cache_path(i_date)
        if os.path.exists(_feedfile):
            break
    else:
        _feedfile = None

    for i_date in reversed(list(utilities.iterate_dates(1995, period='quarterly'))):
        _idxfile = config.get_index_cache_path(i_date)
        if os.path.exists(_idxfile):
            break
    else:
        _idxfile = None

    print("Last downloaded feed cache: ", _feedfile)
    print("Last downloaded index cache:", _idxfile)

def print_config():
    """Prints out config file"""
    clean_user = lambda s: s.replace(os.path.expanduser("~"), "~") if isinstance(s, str) else s
    _cfloc = config.get_config_file()

    print("pyEDGAR Config File:")
    if _cfloc is None:
        print("No config file found in:")
        for x in config.PREFERRED_CONFIG_DIRECTORIES:
            if x == os.path.abspath("."):
                print("\t[current directory]:", end="")
            print("\t{}/".format(clean_user(x).replace("\\", "/")))
    else:
        print("Location: {}".format(clean_user(_cfloc)))

    print("-" * 80)
    _config = {c: getattr(config, c) for c in dir(config) if c.isupper()}
    for c in "PREFERRED_CONFIG_DIRECTORIES CONFIG_OBJECT".split():
        if c in _config:
            del _config[c]

    _maxlen = max(len(k) for k in _config.keys())
    for k, v in _config.items():
        print("{1:>{0}s}: {2}".format(_maxlen, k, clean_user(v)))
    print("-" * 80)


if __name__ == "__main__":
    from argparse import ArgumentParser

    argp = ArgumentParser(
        description="Downloader for pyedgar. Has functionality for downloading indexes and filings.\n"
        "Indexes: downloads quarterly index files from EDGAR, then combines them all into one big master index file.\n"
        "Filings: downloads compressed filing 'feed files' from EDGAR, then extracts the specified form types."
    )

    argp.add_argument(
        "-s",
        "--start-date",
        default=None,
        dest="start_date",
        metavar="YYYY-MM-DD",
        type=utilities.parse_date_input,
        help="An optional date to start downloading feeds/indices from (of form YYYY-MM-DD).",
    )

    argp.add_argument(
        "-n",
        "--last-n-days",
        default=30,
        dest="last_n_days",
        type=int,
        help="An number of days before today to start downloading feeds/indices from.",
    )

    argp.add_argument("-i", "--indices", action="store_true", dest="get_indices", help="Download and update indices.")

    argp.add_argument(
        "-d", "--download-feeds", action="store_true", dest="get_feeds", help="Download and extract daily feed feeds.",
    )

    argp.add_argument("--config", action="store_true", dest="print_config", help="Print config file settings.")

    argp.add_argument("--status", action="store_true", dest="print_status", help="Show last downloaded feed and index cache files.")

    argp.add_argument(
        "--log",
        "--log-level",
        dest="log_level",
        default="error",
        help="Set the log-level to display more/less output. Choose from: error (default), warning, info, debug.",
    )

    cl_args = argp.parse_args()

    _log_level = {"w": logging.WARNING, "i": logging.INFO, "d": logging.DEBUG,}.get(
        cl_args.log_level[0].lower(), logging.ERROR
    )

    logging.basicConfig(level=_log_level)

    _logger.debug("Running with args: %r", cl_args)

    if cl_args.print_status:
        print_cache_status()

    if cl_args.print_config:
        print_config()

    main(
        start_date=cl_args.start_date,
        last_n_days=cl_args.last_n_days,
        get_indices=cl_args.get_indices,
        get_feeds=cl_args.get_feeds,
    )



# -*- coding: utf-8 -*-
"""
Base class for EDGAR filing.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# import os
import re
import logging

from pyedgar import config
from pyedgar.utilities import edgarweb
from pyedgar.utilities import forms
from pyedgar.utilities.forms import FORMS
from pyedgar.utilities import localstore


class Filing(object):
    """
    Base class for EDGAR filing, with format aware parsing
    to provide access to filing documents as objects.
    """

    #: Firm CIK
    _cik = None
    #: Filing Accession
    _accession = None
    #: Whether filings are cached locally, or to get them from EDGAR
    _local_cache = False
    #: If filing is not found, fallback to downloading from the web.
    _web_fallback = True
    #: The 'loose' form type, with some family hierarchy: 10s, Def 14s, etc.
    _type = None
    #: The exact form type, extracted from the main <TYPE> tag.
    _type_exact = None
    #: Paths to local filing. Set lazily.
    _filing_local_path = None
    #: Paths to remote filing. Set lazily.
    _filing_url = None
    #: Form Data (main headers)
    _headers = {}
    #: Full text of the document, loaded lazily.
    _full_text = None
    #: Array of filed exhibits, parsed lazily.
    _documents = None
    #: Object logger for logging things and stuff.
    __log = None
    #: Read-filing arguments
    read_args = None
    #: get_header arguments
    header_args = None

    def __init__(
        self,
        cik=None,
        accession=None,
        use_cache=None,
        web_fallback=True,
        flat_headers=True,
        omit_duplicate_headers=False,
        duplicate_headers_as_list=True,
        **kwargs
    ):
        """
        Initialization sets CIK, Accession, and optionally
        loads filing from disk.

        Args:
            cik (int,str): Numeric CIK for firm.
            accession (str): Accession for filing, filed under CIK.
                Expected format: 0123456789-01-012345, or 012345678901012345.
            use_cache (bool): Use the local cache at `config.FEED_CACHE_ROOT`,
                default to `config.CACHE_FEED`.
            web_fallback (bool): When FileNotFoundError is raised, try
                `edgarweb.download_form_from_web` instead.
            flat_headers (bool): Set the style of loading headers. If True,
                all headers will be loaded in one dictionary. If False, headers
                will be loaded as a hierarchical set of dictionaries, matching
                the hierarchy in the filing's headers. Default: True.
            omit_duplicate_headers (bool): Passed to `get_all_headers()`, if True
                will ignore duplicate headers (e.g. 8-K items) keeping the first.
                Default: False.
            duplicate_headers_as_list (bool): Passed to `get_all_headers()`,
                if True will return the header values as a list (e.g. ['5.02',
                '5.07', '9.01']). If False will add _# to duplicate header names.
                Default: True.

        Returns:
            Filing object.

        Raises:
            None. Loading done lazily.
        """
        self.__log = logging.getLogger("pyedgar.filing.Filing")

        self._set_cik(cik)
        self._set_accession(accession)

        self._local_cache = use_cache if use_cache is not None else config.CACHE_FEED
        self._web_fallback = web_fallback

        self.read_args = kwargs
        self.header_args = {
            "flat": flat_headers,
            "omit_duplicates": omit_duplicate_headers,
            "add_int_to_name": not duplicate_headers_as_list,
        }

    def __repr__(self):
        return "<EDGAR filing ({}/{}) Headers:{}, Text:{}, Documents:{}>".format(
            self.cik, self.accession, bool(self._headers), bool(self._full_text), bool(self._documents),
        )

    def __str__(self):
        return self.__repr__()

    def _set_cik(self, cik=None):
        """
        Set cik on object, verifying format is CIK-like.

        Args:
            cik (int,str): Numeric CIK for firm, either as int or string.

        Returns:
            int: CIK of the filing.

        Raises:
            ValueError: if cik is not numerical (or castable to int).
        """
        try:
            if cik is not None:
                self._cik = int(cik)
        except ValueError:
            # They didn't pass in a CIK that looked like a number
            raise ValueError("CIKs must be numeric variables," " you passed in {}".format(cik))

        return self._cik

    def _set_accession(self, accession=None):
        """
        Set accession on object, verifying format is accession-like.
        Expected format: 0123456789-01-012345, or 012345678901012345.
        Function converts the latter to the former.

        Args:
            accession (str): Accession number as a string, with format
                0123456789-01-012345, or 012345678901012345.

        Returns:
            The formatted Accession as a string of 20 characters.

        Raises:
            ValueError: if accession is not expected format.
        """
        try:
            if accession and localstore.ACCESSION_RE.search(accession):
                if len(accession) == 18:
                    self._accession = "{}-{}-{}".format(accession[:10], accession[10:12], accession[12:])
                else:
                    self._accession = accession
        except TypeError:
            # They didn't pass in an accession that was a string.
            raise ValueError(
                "Accessions must be 18/20 character strings of format"
                " ##########-##-######, you passed in {}".format(accession)
            )

        return self.accession

    def _set_full_text(self):
        """
        Full text of the filing at cik/accession.
        Lazily load the full text of the filing into memory.
        Use cache if `self._local_cache`, and fall back to EDGAR website
        if `self._web_fallback`.

        Returns:
            String representing the full text of the EDGAR filing.

        Raises:
            FileNotFoundError: The file wasn't found in the local cache if
                `self._local_cache`.
        """
        if not self._full_text:
            self.__log.debug("Local cache: %r", self._local_cache)
            if self._local_cache:
                self.__log.debug("Local cache if: %r", self._local_cache)
                try:
                    self._full_text = forms.get_full_filing(self.path, **self.read_args)
                    return self._full_text
                except FileNotFoundError:
                    msg = "Filing not found for CIK:{} / Accession:{}".format(self.cik, self.accession)
                    self.__log.debug(msg)
                    if not self._web_fallback:
                        raise FileNotFoundError(msg)

            self.__log.debug("Downloading from EDGAR web: %d/%s", self.cik, self.accession)
            self._full_text = edgarweb.download_form_from_web(self.cik, self.accession)

        return self._full_text

    def _set_headers(self, **load_kwargs):
        """
        Load the full set of headers of the filing at cik/accession into memory.

        Args:
            Optional load arguments passed to `forms.get_all_headers()`

        Returns:
            dict: Dictionary of headers, with either flat, hierarchical, both,
                or neither, depending on `self._flat_headers`.
        """
        if not self.full_text:
            self.__log.debug("Full filing text missing or not found!")
            return None

        self._headers = forms.get_all_headers(self.full_text, **{**self.header_args, **load_kwargs})

        return self._headers

    def _set_type(self):
        """
        Full text of the filing at cik/accession.
        Lazily load the full text of the filing into memory.

        Returns:
            String representing the full text of the EDGAR filing.

        Raises
            ileNotFoundError: The file wasn't found in the local cache.
        """
        _t = self.headers.get("type", self.headers.get("conformed-submission-type", "OTHER"))
        self._type_exact = _t

        self._type = "other"
        if _t in ("3", "3/A"):
            self._type = FORMS.FORM_3
        elif _t in ("4", "4/A"):
            self._type = FORMS.FORM_4
        elif _t in ("8-K", "8-K/A"):
            self._type = FORMS.FORM_8K
        elif _t[:4] in ("10-Q" "10QS"):
            self._type = FORMS.FORM_10Q
        elif _t[:4] in ("10-K" "10KS"):
            self._type = FORMS.FORM_10K
        elif _t.endswith("14A"):
            self._type = FORMS.FORM_DEF14A
        elif "SC 13G" in _t:
            self._type = FORMS.FORM_13G
        elif "SC 13D" in _t:
            self._type = FORMS.FORM_13D
        elif "13F-" in _t:
            self._type = FORMS.FORM_13F

        return self._type

    def _set_documents(self):
        """
        Parse the full text of the filing and split it into the
        documents/exhibits with associated meta data.
        Full text of the documents resides at documents[i]['full_text'].

        Returns:
            list: List of document objects. ['full_text'] contains the document
                text.
        """
        if not self.full_text:
            self.__log.debug("Full filing text missing or not found!")
            return None
        self._documents = forms.chunk_filing(self.full_text)

        return self._documents

    cik = property(fget=lambda self: self._cik, fset=_set_cik)
    accession = property(fget=lambda self: self._accession, fset=_set_accession)

    @property
    def path(self):
        """
        Get the filing path internal variable.
        Sets it lazily the first time it is used.

        Returns:
            str: The full path of the local cache.
        """
        if not self._filing_local_path:
            self._filing_local_path = localstore.get_filing_path(cik=self.cik, accession=self.accession)

        return self._filing_local_path

    @property
    def urls(self):
        """
        Get the URLs to the filing path.

        Returns:
            tuple: (url to raw filing, url to index on EDGAR)
        """
        if not self._filing_url:
            self._filing_url = edgarweb.get_edgar_urls(self.cik, self.accession)

        return self._filing_url

    @property
    def full_text(self):
        """
        Full text of the filing at cik/accession.
        Lazily load the full text of the filing into memory.

        Returns:
            str: Full document of the filing, equivalent to viewing raw file.
        """
        return self._full_text or self._set_full_text()

    @property
    def headers(self):
        """
        Full set of headers of the filing at cik/accession.
        Lazily load the headers of the filing into memory.

        Returns:
            dict: Dictionary of the headers, either flat or hierarchical
                depending on the filing object attributes.
        """
        return self._headers or self._set_headers()

    @property
    def type(self):
        """
        Generic type of the filing at cik/accession, from:
        3, 4, 8-K, 10-K, 10-Q, DEF14A, 13G, 13D, 13F
        Lazily load the headers of the filing into memory.

        Returns:
            str: Type string of the document, simplified.
        """
        return self._type or self._set_type()

    @property
    def type_exact(self):
        """
        Exact TYPE header of the filing at cik/accession.
        Lazily load the headers of the filing into memory.

        Returns:
            str: Full type string of the document from the header.
        """
        return self._type or self._set_type()

    @property
    def documents(self):
        """
        Full set of documents/exhibits of the filing at cik/accession.
        Lazily load the documents of the filing into memory.
        Full text of the documents resides at documents[i]['full_text'].

        Returns:
            list: List of all documents in the filing.
        """
        return self._documents or self._set_documents()

    def get_sequence_number(self, sequence_number):
        """
        Access exhibits (or main filing) by sequence number (1-indexed).

        Args:
            sequence_number (int): 1-indexed int representing the sequence
                of the document in the EDGAR filing. Identified using
                the 'sequence' tag in each <DOCUMENT> element.

        Returns:
            Array of document dictionaries in the form:
            ``[{search_key: search_string, 'full_text':...}, ...]``
            or ``None``, if no document is available or sequence is not found.
        """
        if not self.documents:
            return None

        for doc in self.documents:
            try:
                i_seq = int(doc.get("sequence", -1337))
                if sequence_number == i_seq:
                    return doc
            except ValueError:
                # then try matching sequences as strings.
                i_seq = doc.get("sequence", "").strip()
                if str(sequence_number) == i_seq:
                    return doc

    def get_documents_by_tag(self, tag_name, search_string, regex=False, flags=0):
        """
        Search document dictionaries for `tag_name` that matches `search_string`.
        Returns array of document dictionaries that match the requirement:
        ``search_string == document[tag_name]``, or using `regex=True`,
        ``search_string.search(document['filename'])``
\
        Args:
            tag_name (str): dictionary key in document dicts to match against.
            search_string (str): string to search for in ``document[tag_name]``
            regex (str): Boolean signifying that search_string should use a regular
                expression match. If search_string isn't a compiled regex
                pattern, then compile it using flags=flags.
            flags (str): Optional flags to use when compiling search_string,
                if it isn't a ``re.Pattern`` already.

        Returns:
            Array of document dictionaries in the form:
            ``[{tag_name: search_string, 'full_text':...}, ...]``
            or ``None``, if no document is available or string is not found.
        """
        if not self.documents or not search_string or not tag_name:
            return None

        if regex:
            search_re = search_string
            if not hasattr(search_re, "search"):
                search_re = re.compile(search_re, flags=flags)

        ret = []
        for doc in self.documents:
            if tag_name not in doc:
                continue
            match_against = doc.get(tag_name)
            if not regex and search_string == match_against:
                ret.append(doc)
            elif regex and search_re.search(match_against):
                ret.append(doc)

        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Download and cache a local copy of EDGAR.
Allows for caching of downloaded daily zips and index files.

EDGAR HTML specification: https://www.sec.gov/info/edgar/ednews/edhtml.htm
EDGAR FTP specification: https://www.sec.gov/edgar/searchedgar/ftpusers.htm

URL Change in 2016:
  <2016 ftp URL: ftp://ftp.sec.gov/edgar/data/2098/0000002098-96-000003.txt
  >2016 http URL: https://www.sec.gov/Archives/edgar/data/2098/0000002098-96-000003.txt

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import os
import re
import tarfile
import logging

# 3rd party imports

# Module Imports
from pyedgar.exceptions import InputTypeError, WrongFormType, NoFormTypeFound, NoCIKFound
from pyedgar import config
from pyedgar import utilities
from pyedgar.utilities import localstore
from pyedgar.utilities import forms
from pyedgar.utilities import edgarweb


class EDGARCacher(object):
    """
    Class that downloads EDGAR to your very own computer.
    Everyone should have a local copy after all!
    I have just enough flexibility here for my computer. It works on Linux, YWindowsMMV.
    """

    # These should work for everypeople.
    EDGAR_ENCODING = "latin-1"  # The SEC documentation says it uses latin-1
    # EDGAR_ENCODING = 'utf-8' # Alternatively one could use utf-8.

    # These should be changed for you. Either in source, or more better at runtime.
    keep_regex = None
    check_cik = False

    # Class local vars
    _path_formatter = None
    _use_requests = False

    # Local versions of file path lookups, for overriding if you like
    _get_filing_path = None
    _get_feed_cache_path = None
    _get_index_cache_path = None

    # May as well share this across instances (instead of setting in __init__)
    _logger = logging.getLogger(__name__)

    def __init__(self, keep_form_type_regex=None, check_cik=False, use_tqdm=True, use_requests=False):
        """
        Initialize the downloader object.

        keep_form_type_regex: regular expression object which will match to form-types you wish to keep, default to config file if None. Pass '.' for all filings.
        check_cik: flag for whether to extract the CIK from the nc file for passing into the path formatter.
        use_tqdm: flag for whether or not to wrap downloads in tqdm for progress monitoring
        """
        self.check_cik = check_cik
        self._use_requests = use_requests

        self.keep_regex = keep_form_type_regex
        if keep_form_type_regex is None and config.KEEP_REGEX:
            self.keep_regex = re.compile(config.KEEP_REGEX)

        self._get_filing_path = localstore.get_filing_path
        self._get_feed_cache_path = config.get_feed_cache_path
        self._get_index_cache_path = config.get_index_cache_path

        self._logger.debug("Feed cache: %r | %r", config.FEED_CACHE_ROOT, config.FEED_CACHE_PATH_FORMAT)
        self._logger.debug("Filing root: %r | %r", config.FILING_ROOT, config.FILING_PATH_FORMAT)
        self._logger.debug("Index cache: %r | %r", config.INDEX_CACHE_ROOT, config.INDEX_CACHE_PATH_FORMAT)
        self._logger.debug("Index root: %r | form_*.%s", config.INDEX_ROOT, config.INDEX_EXTENSION)
        self._logger.debug("Extract regex: %r", self.keep_regex.pattern if self.keep_regex else "ALL")

    def _handle_nc(self, file_or_str):
        """
        Reads file or string, returns dictionary based on flags or None on failure.
        At a minimum, it returns: {'doc'}

        If keep_regex is set (not None), it adds 'form_type'.
            (Raises WrongFormType if FORM-TYPE SGML tag doesn't match `keep_regex`.)

        If check_cik is True, it extracts and adds 'cik'.
        """
        try:
            txt = file_or_str.read()
        except AttributeError:
            txt = file_or_str

        if not txt:
            raise InputTypeError("No text of file object found")

        for _decode_type, _errors in zip((self.EDGAR_ENCODING, "utf-8", "latin-1"), ("strict", "strict", "ignore")):
            try:
                txt = txt.decode(_decode_type, errors=_errors)
                break
            except (UnicodeDecodeError, ValueError):
                pass

        ret_val = {"doc": txt, "encoding": _decode_type, "decode_errors": _errors}

        if self.keep_regex is not None:
            ret_val["form_type"] = forms.get_header(txt, "FORM-TYPE")

            if not ret_val["form_type"]:
                raise NoFormTypeFound(ret_val["form_type"])

            if not self.keep_regex.search(ret_val["form_type"]):
                raise WrongFormType(ret_val["form_type"])

        if self.check_cik:
            ret_val["cik"] = forms.get_header(txt, "CIK")
            ret_val["accession"] = forms.get_header(txt, "ACCESSION-NUMBER")

            if not ret_val["cik"]:
                raise NoCIKFound("No CIK found in {}".format(txt[:250]))

        return ret_val

    def extract_from_feed_cache(self, cache_path, overwrite=False):
        """
        Extract all filings from a daily feed compressed cache file.
        Uses `self._get_filing_path` to determine location for extracted filings.

        Extracts filings based on regular expression match of form type if `self.keep_regex` is not None.
        """
        i_done, i_tot = 0, 0

        with tarfile.open(cache_path, "r") as tar:
            for tarinfo in tar:
                if len(tarinfo.name) < 3 or ".corr" in tarinfo.name:
                    continue
                i_tot += 1

                # tarinfo.name of form ./ACCESSION.nc
                nc_acc = tarinfo.name.split("/")[-1][:-3]
                if len(nc_acc) != 20:
                    self._logger.warning("\tAccession in filename seems suspect. %r", nc_acc)

                try:
                    nc_file = tar.extractfile(tarinfo)
                except IOError:
                    continue

                # At this point, we have a file, and we have an accession.
                # This should be all we need to save it off.
                try:
                    nc_dict = self._handle_nc(nc_file)
                except InputTypeError:
                    self._logger.warning("\tNot a file or string at %r (%r/%r extracted)", tarinfo.name, i_done, i_tot)
                    continue
                except NoCIKFound:
                    # This only triggers if self.check_cik is set.
                    self._logger.warning("\tNo CIK found at %r (%r/%r extracted)", tarinfo.name, i_done, i_tot)
                    continue
                except NoFormTypeFound:
                    # This only triggers if self.keep_regex is set.
                    self._logger.warning("\tNo FormType found at %r (%r/%r extracted)", tarinfo.name, i_done, i_tot)
                    continue
                except WrongFormType:
                    # This only triggers if self.keep_regex is set.
                    continue

                try:
                    nc_text = nc_dict.pop("doc")
                    if "accession" not in nc_dict:
                        nc_dict["accession"] = nc_acc
                except AttributeError:
                    # None type has no pop
                    self._logger.warning(
                        "\tHandling nc file %r passed exceptions (%r/%r extracted)", tarinfo.name, i_done, i_tot
                    )
                    continue
                except KeyError:
                    # This triggers upon nc_dict.pop not having 'doc' in it. Shouldn't happen.
                    self._logger.warning(
                        "\tNo document item extracted from %r (%r/%r extracted)", tarinfo.name, i_done, i_tot
                    )
                    continue

                # Get local nc file path. Accession is nc file filename.
                nc_out_path = self._get_filing_path(**nc_dict)
                i_done += 1
                if not overwrite and os.path.exists(nc_out_path):
                    continue

                # Sometimes the containing dir (cik or year) doesn't exist. Make it so.
                if not os.path.exists(os.path.dirname(nc_out_path)):
                    os.makedirs(os.path.dirname(nc_out_path))

                with open(nc_out_path, "w", encoding=nc_dict["encoding"], errors=nc_dict["decode_errors"]) as fh:
                    fh.write(nc_text)

        return i_done, i_tot

    def extract_daily_feeds(self, from_date, to_date=None, download_first=False, overwrite=False):
        """
        Loop through daily feed compressed files and extract them to local cache.
        Uses the object's path formatter to determine file paths.

        Args:
            from_date (datetime): Day to start extracting on.
            to_date (datetime): Optional day to finish extracting on. Default: datetime.date.today()
            download_first (bool): Flag for whether to try and download daily feed cache files from
                EDGAR if they don't exist, or just use the already downloaded files. Default: False.
            overwrite (bool): Flag for whether to overwrite filings if they have already been extracted. Default: False.
        """
        num_extracted, num_total, num_parsed, i_extracted, i_searched = 0, 0, 0, 0, 0

        for i_date in utilities.iterate_dates(from_date, to_date=to_date, period="daily"):
            if download_first:
                feed_path = edgarweb.download_feed(i_date, overwrite=overwrite, use_requests=self._use_requests)
            else:
                feed_path = self._get_feed_cache_path(i_date)

            if feed_path is None or not os.path.exists(feed_path):
                self._logger.info("Cache file does not exist for %s at %s.", i_date, feed_path)
                continue

            self._logger.info(
                "%s feed file %s: %4.2f MB",
                "Downloaded" if download_first else "Parsing",
                i_date,
                os.path.getsize(feed_path) / 1024 ** 2,
            )

            try:
                i_extracted, i_searched = self.extract_from_feed_cache(feed_path, overwrite=overwrite)
            except tarfile.ReadError:
                self._logger.error("Handling nc file on %s at %s raised Read Error", i_date, feed_path)

            # Log progress after each tar file is done
            self._logger.info("Extracted %d out of %d filings from %s", i_extracted, i_searched, feed_path)

            num_extracted += i_extracted
            num_total += i_searched
            num_parsed += 1

        return num_extracted, num_total, num_parsed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Download and extract indices from EDGAR index files.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import os
import logging

# 3rd party imports
import pandas as pd

# Module Imports
from pyedgar import config
from pyedgar import utilities
from pyedgar.utilities import edgarweb


# progress logging
try:
    from tqdm import tqdm
except ImportError:

    def tqdm(_iterable, *args, **kwargs):
        return _iterable


class IndexMaker:
    """
    Class that downloads EDGAR to your very own computer.
    Everyone should have a local copy after all!
    I have just enough flexibility here for my computer. It works on Linux, YWindowsMMV.
    """

    edgar_index_args = {"sep": "|", "encoding": "latin-1", "skiprows": [0, 1, 2, 3, 4, 5, 6, 7, 8, 10]}

    _use_requests = False
    _tq = None
    _get_index_cache_path = None

    # May as well share this across instances (instead of setting in __init__)
    _logger = logging.getLogger(__name__)

    def __init__(self, use_tqdm=False, use_requests=False):
        """
        Initialize the index making object.

        Arguments:
            use_tqdm (bool): flag for whether or not to wrap downloads in tqdm for progress monitoring
        """
        self._use_requests = use_requests
        self._get_index_cache_path = config.get_index_cache_path

        if use_tqdm:
            self._tq = tqdm
        else:

            def _tq(_iterable, *args, **kwargs):
                return _iterable

            self._tq = _tq

    def download_indexes(self, start_year=1995, stop_year=None, overwrite=False):
        """Download multiple edgar quarterly index compressed files.

        Args:
            start_year (int, datetime, None): Starting year (or datetime from which we'll extract year/quarter). Default to 1995.
            stop_year (int, datetime, None): Ending year (or datetime from which we'll extract year/quarter). Default to today's year.
            overwrite (bool): Flag for whether to overwrite any existing file (default False).
            use_requests (bool): Flag for whether to use requests or curl (default False == curl).
            overwrite_size_threshold (int): Existing files smaller than this will be re-downloaded.

        Returns:
            tuple: output file path, return code
        """
        _num = len([0 for _ in utilities.iterate_dates(start_year, stop_year, period="quarterly")])

        # The download recursively works as it sounds, but we want progress bar, so do it date by date
        for i_date in self._tq(
            utilities.iterate_dates(start_year, stop_year, period="quarterly"), total=_num, desc="Downloading Indices"
        ):
            edgarweb.download_indexes_recursively(
                i_date, end_date=i_date, overwrite=overwrite, use_requests=self._use_requests
            )

    def extract_indexes(self, start_year=1995, stop_year=None, save_forms=None, download_first=True, overwrite=False):
        if download_first:
            self._logger.info("Downloading the quarterly indices...")
            self.download_indexes(start_year=start_year, stop_year=stop_year, overwrite=overwrite)
            self._logger.info("Done downloading quarterly indices.")

        df = pd.DataFrame()

        _num = len([0 for _ in utilities.iterate_dates(start_year, stop_year, period="quarterly")])

        for i_date in self._tq(
            utilities.iterate_dates(start_year, to_date=stop_year, period="quarterly"),
            total=_num,
            desc="Extracting Indices",
        ):
            idx_cache_file = self._get_index_cache_path(i_date)

            try:
                self._logger.info("\tLoading index for %rQ%r", i_date.year, utilities.get_quarter(i_date))
                dfi = pd.read_csv(idx_cache_file, **self.edgar_index_args)
            except FileNotFoundError:
                self._logger.warning(
                    "No Index cache file at %r (for %rQ%r)", idx_cache_file, i_date.year, utilities.get_quarter(i_date)
                )
            except Exception:
                self._logger.warning(
                    "Reading %r failed at %rQ%r", idx_cache_file, i_date.year, utilities.get_quarter(i_date)
                )
                # File reading didn't work, try overwriting with new download
                continue

            dfi["Accession"] = dfi.Filename.str.slice(start=-24, stop=-4)
            del dfi["Filename"]

            df = pd.concat([df, dfi], copy=False)

            self._logger.info("Added %r(%r) to df(%r)", idx_cache_file, len(dfi), len(df))

        df["Date Filed"] = pd.to_datetime(df["Date Filed"])

        all_forms = df["Form Type"].unique()
        if save_forms is None:
            save_forms = {
                "all": all_forms,
                "10-K": [x for x in all_forms if x[:4] in ("10-K", "10KS")],
                "10-Q": [x for x in all_forms if x[:4] in ("10-Q", "10QS")],
                "DEF14A": [x for x in all_forms if x.endswith("14A")],
                "8-K": ("8-K", "8-K/A"),
            }

        for form, formlist in self._tq(save_forms.items(), total=len(save_forms), desc="Exporting Indices"):
            outpath = os.path.join(config.INDEX_ROOT, "form_{}.{}".format(form, config.INDEX_EXTENSION))

            self._logger.info("Saving %r to %r", form, outpath)
            self._logger.debug("Saving %s to %s with form-types: %r", form, outpath, formlist)

            (
                df[df["Form Type"].isin(formlist)]
                .sort_values(["CIK", "Date Filed"])
                .to_csv(outpath, sep=config.INDEX_DELIMITER, index=False)
            )
        self._logger.info("Done extracting indices!")



# -*- coding: utf-8 -*-
"""
Utilities for interacting with edgar forms.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

import re
import os
import logging

# from . import plaintext
from .htmlparse import convert_html_to_text
from ..exceptions import WrongFormType, EDGARFilingFormatError

__logger = logging.getLogger(__name__)

# Obviously encoding is not uniform. Because internet.
# ENCODING_INPUT = 'latin-1'
ENCODING_INPUT = "utf-8"
ENCODING_OUTPUT = "utf-8"

RE_DOC_TAG = re.compile("</?DOCUMENT>")
RE_DOC_TAG_OPEN = re.compile("<DOCUMENT>")
RE_DOC_TAG_CLOSE = re.compile("</DOCUMENT>")
RE_TEXT_TAG = re.compile("</?TEXT>")
RE_TEXT_TAG_OPEN = re.compile("<TEXT>")
RE_TEXT_TAG_CLOSE = re.compile("</TEXT>")
# Only matches <KEY>VALUE, no </KEY> nor <KEY>\n
RE_HEADER_TAG = re.compile(r"^<(?P<key>[^/][^>]*)>[ \t]*(?P<value>.+)$", re.M)
# Matches anything like <*>*\n
RE_HEADER_TAG_OC = re.compile(r"^<(?P<key>/?[^>]*)>(?P<value>.*)$", re.M)
# Matches KEY: value
RE_HEADER_TAG_PLAINTEXT = re.compile(r"^(?P<indent>[ \t]*)(?P<key>[^\n\r:]+):[ \t]*(?P<value>[^\n\r]+)?$", re.M)


def get_full_filing(file_path, encoding=None, errors="ignore"):
    """
    Returns full text of filing.
    Calls: `open(file_path, encoding=encoding or ENCODING_INPUT, errors=errors)`

    Args:
        file_path (str or Path): path to the filing to be loaded.
        encoding (str): encoding to use in reading file. Default: `ENCODING_INPUT` (utf-8)
        errors (str): how to handle errors, passed to `open`. Default: `'ignore'`.

    Returns:
        str: Full text of the filing

    Raises:
        FileNotFoundError: Raised if file doesn't exist.
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError("File {} does not exist.".format(file_path))

    with open(file_path, encoding=encoding or ENCODING_INPUT, errors=errors or "ignore") as fh:
        return fh.read()


def get_form_with_header(file_path, form_type=None, buff_size=(2 << 16) + 8, encoding=None, errors="ignore"):
    """
    Reads in filing located at `file_path` and returns dictionary of its
    contents and some header information.

    Args:
        file_path (str or Path): path to the filing to be loaded.
        form_type (str): Verify filings TYPE header matches `form_type` or raise `WrongFormType`.
        buff_size (int): Buffer size to read in filings. Default: (2<<16) + 8
        encoding (str): encoding to use in reading file. Default: `ENCODING_INPUT` (utf-8)
        errors (str): how to handle errors, passed to `open`. Default: `'ignore'`.

    Returns:
        dict: Filing and information of the form: {'cik', 'form_type', 'filing_date', 'text':[]}

    Raises:
        FileNotFoundError: Raised if file doesn't exist.
        WrongFormType: Raised if form TYPE header doesn't match `form_type`
        EDGARFilingFormatError: Raised if filing is misformatted (no starting `<DOCUMENT>` tag)
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError("File {} does not exist.".format(file_path))

    with open(file_path, encoding=encoding or ENCODING_INPUT, errors=errors or "ignore", buffering=buff_size) as fh:
        text = fh.read(buff_size)

        found_form = get_header(text, "TYPE")
        if form_type is not None:
            if not found_form or form_type.upper() != found_form.upper():
                raise WrongFormType

        # Now find where the header stops (where first document starts)
        doc_start = RE_DOC_TAG_OPEN.search(text)

        # If no DOCUMENT tag found, this isn't an EDGAR form. ABORT!
        if not doc_start:
            raise EDGARFilingFormatError
        # This is what I care about now. Could be changed to `get_all_headers`
        ret_dict = {
            "form_type": found_form.upper(),
            "name": get_header(text, "CONFORMED-NAME", endpos=doc_start.start()),
            "sic": get_header(text, "ASSIGNED-SIC", endpos=doc_start.start()),
            "fye": get_header(text, "FISCAL-YEAR-END", endpos=doc_start.start()),
            "filing_date": get_header(text, "FILING-DATE", endpos=doc_start.start()),
            "filing_date_period": get_header(text, "PERIOD", endpos=doc_start.start()),
            "filing_date_change": get_header(text, "DATE-OF-FILING-DATE-CHANGE", endpos=doc_start.start()),
        }
        # Iteratively loop through open file buffer, reading buff_size chunks
        # until </DOCUMENT> tag is found. There is a chance that the tag could
        # be split across chunks, but it's a cost I'm willing to accept.
        chunks = [text]
        while not RE_DOC_TAG_CLOSE.search(chunks[-1]):
            text = fh.read(buff_size)
            if not text:  # prevent infinite loop, text is null when EOF reached
                break
            chunks.append(text)

    # Now put all those chunks together.
    text = "".join(chunks)
    st = RE_DOC_TAG_OPEN.search(text)
    if not st:
        ret_dict["text"] = text
        return ret_dict
    en = RE_DOC_TAG_CLOSE.search(text, st.end())  # start searching after start
    if not en:
        ret_dict["text"] = text[st.end()]
        return ret_dict

    ret_dict["text"] = text[st.end() : en.start()]

    return ret_dict


def get_form(file_path, encoding=None, errors="ignore", chunk_size=1024 ** 2):
    """
    Reads file at file_path and returns form between <TEXT> and </TEXT> tags.
    Default is to chunk the file being read, because sometimes 225MB files have
    only 1MB of text in the first document. So speed that up.
    Set chunk_size=None to disable chunking.

    Args:
        file_path (str or Path): path to the filing to be loaded.
        encoding (str): encoding to use in reading file. Default: `ENCODING_INPUT` (utf-8)
        errors (str): how to handle errors, passed to `open`. Default: `'ignore'`.
        chunk_size (int): Size of chunks to read of file, until end of first document
            is found. Disable with `chunk_size=None`. Default: 1MB

    Returns:
        str: Text of the first document.

    Raises:
        FileNotFoundError: Raised if file doesn't exist.
        EDGARFilingFormatError: Raised if filing is misformatted (no starting `<DOCUMENT>` tag)
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError("File {} does not exist.".format(file_path))

    if chunk_size is not None and os.stat(file_path).st_size < 2 * chunk_size:
        chunk_size = None

    with open(file_path, encoding=encoding or ENCODING_INPUT, errors=errors, buffering=chunk_size or -1) as fh:
        if chunk_size is None:
            text = fh.read()
        else:
            chunks = []
            # Avoid while True, so set range to 10GB / buffer size
            for i in range(int(10 * 1024 ** 3 // chunk_size) + 1):
                chunks.append(fh.read(chunk_size))

                _doc_match = RE_DOC_TAG_CLOSE.search(chunks[-1])
                _txt_match = RE_TEXT_TAG_CLOSE.search(chunks[-1])

                # Look for doc or text end if neither found, continue
                if not _doc_match and not _txt_match:
                    continue

                # If we're here, one or both were found. So we're done.
                break

            text = "".join(chunks)

    en = RE_TEXT_TAG_CLOSE.search(text)  # start searching after start
    if not en:
        raise EDGARFilingFormatError("No ending TEXT tag found, despite ending DOCUMENT found.")

    st = RE_TEXT_TAG_OPEN.search(text, endpos=en.start())
    if not st:
        raise EDGARFilingFormatError("No starting TEXT tag found, despite ending TEXT found.")

    return text[st.end() : en.start()].strip()


def get_plaintext(path, unwrap=True, document_width=150, just_first=True):
    """
    Get the plaintext version of an edgar filing.
    Assumes the first exhibit in the full filing text document.
    If HTML, uses w3m linux program to parse into plain text.
    If `unwrap`, also unwraps paragraphs so each paragraph is on one line.

    Args:
        path (str or Path): Full path to form.
        unwrap (bool): Whether to call `plaintext.unwrap_plaintext` on document.
        document_width (int): How wide the plaintext will be. Used in unwrapping.

    Returns:
        str: Plain text representation of file.
    """
    text = get_form(path)

    return convert_html_to_text(text, unwrap=unwrap, document_width=document_width)


def get_all_headers(text, flat=False, force_sgml=False, **kwargs):
    """
    <KEY>VALUE formatted headers in EDGAR documents.
    Dictionary is either iterative (`flat=False`, default) where open/close
    SGML tags are nested dictionaries, or flat, where multiple keys
    are either ignored, renamed, or put in a list (see `get_all_headers_flat`).

    Note: this works on either SGML headers or formatted, web version headers,
    but SGML parsing can be forced with `force_sgml`.

    Dictionary keys are lowercase (`.lower()` is called), and stripped.
    """
    if "ACCESSION NUMBER:" in text[:2000] and not force_sgml:
        if flat:
            return get_all_headers_flat_nosgml(text, **kwargs)
        return get_all_headers_dict_nosgml(text, **kwargs)
    if flat:
        return get_all_headers_flat(text, **kwargs)
    return get_all_headers_dict(text, **kwargs)


def get_all_headers_flat(text, pos=0, endpos=None, omit_duplicates=False, add_int_to_name=False, **kwargs):
    """
    Return dictionary of all <KEY>VALUE formatted headers in EDGAR documents.
    Note this requires the daily feed version of the EDGAR files.
    Dictionary keys are lowercase (`.lower()` is called), and stripped.

    `pos` and `endpos` can be used to get headers for specific exhibits.

    Keys with the same name can either be ignored (`omit_duplicates=True`),
    or subsequent keys with names already in dictionary can be renamed
    by adding a count to the end (filer, filer_1, etc.) if
    `add_int_to_name=True`, or added to a list if `add_int_to_name=False`.

    Returns dictionary of headers.
    """
    if endpos is None:
        _, endpos = _get_header_bounds(text, pos=pos)

    ret = {}
    for rx in RE_HEADER_TAG.finditer(text, pos, endpos):
        key, val = rx.groups()
        key, val = key.lower().strip(), val.strip()

        if not val:
            continue

        if key not in ret:
            ret[key] = val
            continue

        if omit_duplicates:
            continue

        if add_int_to_name:
            for i in range(1, 10000):
                newkey = "{}_{}".format(key, i)
                if newkey not in ret:
                    key = newkey
                    break
            ret[key] = val
            continue

        if not isinstance(ret[key], list):
            ret[key] = [
                ret[key],
            ]

        ret[key].append(val)

    return ret


def get_all_headers_flat_nosgml(text, pos=0, endpos=None, omit_duplicates=False, add_int_to_name=False, **kwargs):
    """
    Return dictionary of all <KEY>VALUE formatted headers in EDGAR documents.
    Note this requires the daily feed version of the EDGAR files.
    Dictionary keys are lowercase (`.lower()` is called), and stripped.

    `pos` and `endpos` can be used to get headers for specific exhibits.

    Keys with the same name can either be ignored (`omit_duplicates=True`),
    or subsequent keys with names already in dictionary can be renamed
    by adding a count to the end (filer, filer_1, etc.) if
    `add_int_to_name=True`, or added to a list if `add_int_to_name=False`.

    Returns:
        dict: Dictionary of headers.
    """
    if endpos is None:
        _, endpos = _get_header_bounds(text, pos=pos)

    ret = {}
    for rx in RE_HEADER_TAG_PLAINTEXT.finditer(text, pos, endpos):
        _, key, val = rx.groups()
        try:
            key, val = _clean_plaintext_header_key(key.lower()), val.strip()
        except AttributeError:
            # val.strip() breaks if val is None.
            continue

        if not val:
            continue

        if key not in ret:
            ret[key] = val
            continue

        if omit_duplicates:
            continue

        if add_int_to_name:
            for i in range(1, 10000):
                newkey = "{}_{}".format(key, i)
                if newkey not in ret:
                    key = newkey
                    break
            ret[key] = val
            continue

        if not isinstance(ret[key], list):
            ret[key] = [
                ret[key],
            ]

        ret[key].append(val)

    return ret


def get_all_headers_dict(
    text, pos=0, endpos=None, starter_dict=None, omit_duplicates=False, add_int_to_name=False, **kwargs
):
    """
    Return dictionary of all <KEY>VALUE formatted headers in EDGAR documents.
    Note this requires the daily feed version of the EDGAR files.
    Dictionary keys are lowercase (`.lower()` is called), and stripped.

    `pos` and `endpos` can be used to get headers for specific exhibits.

    Keys with the same name can either be ignored (`omit_duplicates=True`),
    or subsequent keys with names already in dictionary can be renamed
    by adding a count to the end (filer, filer_1, etc.) if
    `add_int_to_name=True`, or added to a list if `add_int_to_name=False`.
    """
    if endpos is None:
        _, endpos = _get_header_bounds(text, pos=pos)

    retdict = {}
    if starter_dict is not None:
        retdict.update(starter_dict)
    # push and pop 'current' dictionary from stack, which works because pointers.
    stack = [
        retdict,
    ]

    for imatch in RE_HEADER_TAG_OC.finditer(text, pos, endpos):
        tmp = imatch.groupdict()
        # if 'key' not in tmp or 'value' not in tmp:
        #     continue
        key, val = tmp["key"].lower(), tmp["value"].strip()

        if "/" in key:
            # Then it's a closing tag. pop the last dict off the stack
            if len(stack) > 1:
                stack.pop()
            # If there's a value... we don't care. That's bad formatting.
            continue
        elif val:
            # Then just a normal key.
            # Add to the dict at the end of the stack, checking for dups
            if key not in stack[-1]:
                stack[-1][key] = val
                continue

            if omit_duplicates:
                continue

            if add_int_to_name:
                for i in range(1, 10000):
                    newkey = "{}_{}".format(key, i)
                    if newkey not in stack[-1]:
                        key = newkey
                        break
                stack[-1][key] = val
                continue

            if not isinstance(stack[-1][key], list):
                stack[-1][key] = [
                    stack[-1][key],
                ]

            stack[-1][key].append(val)
            continue

        # Otherwise this might be a sub-group. Look for the </KEY> tag.
        re_end = re.compile("</{}>".format(key), re.I)
        if re_end.search(text, imatch.end(), endpos):
            # Then we found the end tag. Create dict at KEY if KEY doesn't exist
            if key in stack[-1]:
                # Then try key_X for X from 0 to 200 I guess.
                for key_i in range(200):
                    newkey = "{}_{}".format(key, key_i)
                    if newkey not in stack[-1]:
                        key = newkey
                        break
            # make a new dict at KEY
            stack[-1][key] = {}
            stack.append(stack[-1][key])

    return retdict


def get_all_headers_dict_nosgml(text, pos=0, endpos=None, starter_dict=None, **kwargs):
    """
    Return dictionary of all KEY: VALUE formatted headers in EDGAR documents.
    Note this requires the EDGAR website version of the EDGAR files.
    Dictionary keys are lowercase (`.lower()` is called), and stripped.

    `pos` and `endpos` can be used to get headers for specific exhibits.
    """

    def newkey(key):
        if key in stack[-1]:
            # Then try key_X for X from 0 to 2000 I guess.
            for key_i in range(2000):
                newkey = "{}_{}".format(key, key_i)
                if newkey not in stack[-1]:
                    return newkey
        return key

    if endpos is None:
        _, endpos = _get_header_bounds(text, pos=pos)

    retdict = {}
    if starter_dict is not None:
        retdict.update(starter_dict)
    # push and pop 'current' dictionary from stack, which works because pointers.
    stack = [
        retdict,
    ]

    for imatch in RE_HEADER_TAG_PLAINTEXT.finditer(text, pos, endpos):
        tmp = imatch.groupdict()
        # print(f"\nStarting with {tmp}")
        # print(f"Stack length: {len(stack)}")

        indent = tmp["indent"]  # empty for level 0, 1 \t for each level
        key = _clean_plaintext_header_key(tmp["key"].lower())
        try:
            val = tmp["value"].strip()
        except AttributeError:
            val = None

        if len(indent) + 1 < len(stack) and key:
            # We've found an out-dent, drop off a level.
            # Remove last dict from stack and make a new one if val missing
            # print(f"Found new outdent ({key}): len({len(indent)}) + 1 < {len(stack)}")

            while len(stack) > len(indent) + 1:
                stack.pop()
                # print(f"Stack pop, new len: {len(stack)}")

        if not val:
            # We've found the start of a new indent level.
            # Add key to the dict and push new level to stack
            # print(f"Found new indent ({key}): val is {val}")
            key = newkey(key)
            stack[-1][key] = {}
            stack.append(stack[-1][key])
            continue

        # print(f"Adding key/val {key}:{val}")
        stack[-1][newkey(key)] = val

    return retdict


def get_header(text, header, pos=0, endpos=None, return_match=False, return_multiple=False):
    """
    Searches `text` for header formatted <`header`>VALUE\\n and returns VALUE.strip()
    Note this requires the daily feed version of the EDGAR files.

    `pos` and `endpos` can be used to get headers for specific exhibits.
    """
    if endpos is None:
        _, endpos = _get_header_bounds(text, pos=pos)

    re_tag = re.compile(r"^<{}>(.+)$".format(header), re.M | re.I)

    match = re_tag.search(text, pos, endpos)
    value = match.group(1).strip() if match else ""

    if return_match:
        return value, match
    return value


def _get_header_bounds(text, pos=0, endpos=None, **kwargs):
    """
    Return dictionary of all <KEY>VALUE formatted headers in EDGAR documents.
    Note this requires the daily feed version of the EDGAR files.
    Dictionary keys are lowercase (`.lower()` is called), and stripped.

    `pos` and `endpos` can be used to get headers for specific exhibits.

    Keys with the same name can either be ignored (`omit_duplicates=True`),
    or subsequent keys with names already in dictionary can be renamed
    by adding a count to the end (filer, filer1, etc.) if
    `add_int_to_name=True`, or added to a list if `add_int_to_name=False`.

    Returns dictionary of headers.
    """
    if endpos is None:
        doc_tag_open = RE_DOC_TAG_OPEN.search(text, pos=pos)
        if doc_tag_open:
            endpos = doc_tag_open.start()
    if endpos is None:
        endpos = len(text)

    return pos, endpos


def _clean_plaintext_header_key(
    key, replace_with_dash=re.compile(r" (?![1-9])"), replace_with_empty=re.compile(r" (?=[1-9])")
):
    return replace_with_dash.sub("-", replace_with_empty.sub("", key.strip()))


def chunk_filing(text):
    """
    Separates EDGAR filing into constituent documents and
    their associated metadata.

    Args:
        text: Full text of the EDGAR filing, separated by <DOCUMENT> tags.

    Returns:
        List of dictionaries containins:
            {'full_text': full text of document or exhibit, between <TEXT> tags,
             'meta_data': meta data associated with document
                          (between <DOCUMENT> and <TEXT> tags)}

    Raises:
        ValueError: No text provided (empty).
        EDGARFilingFormatError: Document is not in proper EDGAR SGML format
    """
    if not text.strip():
        raise ValueError("Input text is empty.")

    doc_tags = RE_DOC_TAG.findall(text)

    if len(doc_tags) % 2:
        raise EDGARFilingFormatError("Uneven number of <DOCUMENT> tags found." "Found: {}".format(doc_tags))

    for i, tag in enumerate(doc_tags):
        if (i % 2) ^ ("/" in tag):
            raise EDGARFilingFormatError("<DOCUMENT> tags do not alternate open/close." "Found: {}".format(doc_tags))

    docs = []
    for _start, _end in zip(RE_DOC_TAG_OPEN.finditer(text), RE_DOC_TAG_CLOSE.finditer(text)):

        text_start = RE_TEXT_TAG_OPEN.search(text, _start.end(), _end.start())
        if text_start:
            text_end = RE_TEXT_TAG_CLOSE.search(text, _start.end(), _end.start()) or _end
        else:
            # Document with no text? Whatevs.
            text_start = _end
            text_end = _end

        doc = get_all_headers_dict(text, pos=_start.end(), endpos=text_start.start())

        # Slicing that doesn't make sense (i.e. string[5:3]) returns ''
        # Thus if there's no text tag, full_text will be ''
        # because from above, text_start/end are both _end
        doc["full_text"] = text[text_start.end() : text_end.start()]

        docs.append(doc)

    return docs


class CIKS(object):
    GOOGLE = 1288776
    ALPHABET = 1652044
    AMAZON = 1018724
    FACEBOOK = 1326801
    APPLE = 320193
    GE = 40545
    NIKE = 320187
    FORD = 37996
    WALMART = 104169


class FORMS(object):
    FORM_3 = "3"
    FORM_4 = "4"
    FORM_8K = "8-K"
    FORM_10K = "10-K"
    FORM_10Q = "10-Q"
    FORM_DEF14A = "DEF14A"
    FORM_13G = "13G"
    FORM_13D = "13D"
    FORM_13F = "13F"
